/* Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file http_send.c paraslash's http sender */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <lopsub.h>

#include "server.lsg.h"
#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "server.h"
#include "http.h"
#include "list.h"
#include "send.h"
#include "sched.h"
#include "vss.h"
#include "net.h"
#include "fd.h"
#include "chunk_queue.h"

/** Message sent to clients that do not send a valid get request. */
#define HTTP_ERR_MSG "HTTP/1.0 400 Bad Request\n"

/** The possible states of a client from the server's POV. */
enum http_client_status {
	/** We accepted the connection on the tcp socket. */
	HTTP_CONNECTED,
	/** Successfully received the get request. */
	HTTP_GOT_GET_REQUEST,
	/** Connection is ready for sending audio data. */
	HTTP_STREAMING,
	/** We didn't receive a valid get request. */
	HTTP_INVALID_GET_REQUEST
};

/** For each connected client, a structure of this type is maintained. */
struct private_http_sender_data {
	/** The current state of this client. */
	enum http_client_status status;
};

static struct sender_status http_sender_status, *hss = &http_sender_status;

static int http_send_msg(struct sender_client *sc, const char *msg)
{
	int ret = write_buffer(sc->fd, msg);

	if (ret < 0)
		shutdown_client(sc, hss);
	return ret;
}

static void http_send_ok_msg(struct sender_client *sc)
{
	PARA_INFO_LOG("sending http ok message to fd %d\n", sc->fd);
	http_send_msg(sc, HTTP_OK_MSG);
}

static int http_send_err_msg(struct sender_client *sc)
{
	PARA_NOTICE_LOG("sending bad request message to fd %d\n", sc->fd);
	return http_send_msg(sc, HTTP_ERR_MSG);
}

static void http_shutdown_clients(void)
{
	shutdown_clients(hss);
}

static void http_shutdown(void)
{
	http_shutdown_clients();
	generic_acl_deplete(&hss->acl);
}

static int queue_chunk_or_shutdown(struct sender_client *sc,
		struct sender_status *ss, const char *buf, size_t num_bytes)
{
	int ret = cq_enqueue(sc->cq, buf, num_bytes);
	if (ret < 0)
		shutdown_client(sc, ss);
	return ret;
}

/**
 * Send one chunk of audio data to a connected client.
 *
 * \param sc The client.
 * \param ss The sender.
 * \param current_chunk The number of the chunk to write.
 * \param buf The data to write.
 * \param len The number of bytes of \a buf.
 * \param header_buf The audio file header.
 * \param header_len The number of bytes of \a header_buf.
 *
 * On errors, the client is shut down. If only a part of the buffer could be
 * written, the remainder is put into the chunk queue for that client.
 */
static void http_send_chunk(struct sender_client *sc, struct sender_status *ss,
		long unsigned current_chunk, const char *buf, size_t len,
		const char *header_buf,	size_t header_len)
{
	int ret;

	if (!sc->header_sent && current_chunk) {
		if (header_buf && header_len > 0) {
			ret = queue_chunk_or_shutdown(sc, ss, header_buf, header_len);
			if (ret < 0)
				goto out;
		}
	}
	sc->header_sent = 1;
	ret = send_queued_chunks(sc->fd, sc->cq);
	if (ret < 0) {
		shutdown_client(sc, ss);
		goto out;
	}
	if (!len)
		goto out;
	if (!ret) { /* still data left in the queue */
		ret = queue_chunk_or_shutdown(sc, ss, buf, len);
		goto out;
	}
	ret = xwrite(sc->fd, buf, len);
	if (ret < 0) {
		shutdown_client(sc, ss);
		goto out;
	}
	if (ret != len)
		ret = queue_chunk_or_shutdown(sc, ss, buf + ret, len - ret);
out:
	if (ret < 0)
		PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
}

static void http_send(long unsigned current_chunk,
	__a_unused long unsigned chunks_sent, const char *buf, size_t len,
	const char *header_buf, size_t header_len)
{
	struct sender_client *sc, *tmp;

	list_for_each_entry_safe(sc, tmp, &hss->client_list, node) {
		struct private_http_sender_data *phsd = sc->private_data;

		if (phsd->status == HTTP_STREAMING)
			http_send_chunk(sc, hss, current_chunk, buf, len,
				   header_buf, header_len);
	}
}

static void http_post_select(fd_set *rfds, __a_unused fd_set *wfds)
{
	struct sender_client *sc, *tmp;
	struct private_http_sender_data *phsd;
	int ret;

	list_for_each_entry_safe(sc, tmp, &hss->client_list, node) {
		phsd = sc->private_data;
		switch (phsd->status) {
		case HTTP_STREAMING: /* nothing to do */
			break;
		case HTTP_CONNECTED: /* need to recv get request */
			ret = read_pattern(sc->fd, HTTP_GET_MSG, MAXLINE, rfds);
			if (ret < 0)
				phsd->status = HTTP_INVALID_GET_REQUEST;
			else if (ret > 0) {
				phsd->status = HTTP_GOT_GET_REQUEST;
				PARA_INFO_LOG("received get request\n");
			}
			break;
		case HTTP_GOT_GET_REQUEST: /* need to send ok msg */
			phsd->status = HTTP_STREAMING;
			http_send_ok_msg(sc);
			break;
		case HTTP_INVALID_GET_REQUEST: /* need to send err msg */
			if (http_send_err_msg(sc) >= 0)
				shutdown_client(sc, hss);
			break;
		}
	}
	sc = accept_sender_client(hss, rfds);
	if (!sc)
		return;
	phsd = para_malloc(sizeof(*phsd));
	sc->private_data = phsd;
	phsd->status = HTTP_CONNECTED;
}

static void http_pre_select(int *max_fileno, fd_set *rfds, fd_set *wfds)
{
	struct sender_client *sc, *tmp;

	if (hss->listen_fd < 0)
		return;
	para_fd_set(hss->listen_fd, rfds, max_fileno);
	list_for_each_entry_safe(sc, tmp, &hss->client_list, node) {
		struct private_http_sender_data *phsd = sc->private_data;
		if (phsd->status == HTTP_CONNECTED) /* need to recv get request */
			para_fd_set(sc->fd, rfds, max_fileno);
		if (phsd->status == HTTP_GOT_GET_REQUEST ||
				phsd->status == HTTP_INVALID_GET_REQUEST)
			para_fd_set(sc->fd, wfds, max_fileno);
	}
}

static int http_com_on(__a_unused struct sender_command_data *scd)
{
	return generic_com_on(hss, IPPROTO_TCP);
}

static int http_com_off(__a_unused struct sender_command_data *scd)
{
	generic_com_off(hss);
	return 1;
}

static int http_com_deny(struct sender_command_data *scd)
{
	generic_com_deny(scd, hss);
	return 1;
}

static int http_com_allow(struct sender_command_data *scd)
{
	generic_com_allow(scd, hss);
	return 1;
}

static char *http_status(void)
{
	return generic_sender_status(hss, "http");
}

/*
 * Initialize the client list and the access control list, and optionally
 * listen on the tcp port.
 */
static void http_send_init(void)
{
	int ret;

	init_sender_status(hss, OPT_RESULT(HTTP_ACCESS),
		OPT_UINT32_VAL(HTTP_PORT), OPT_UINT32_VAL(HTTP_MAX_CLIENTS),
		OPT_GIVEN(HTTP_DEFAULT_DENY));
	if (OPT_GIVEN(HTTP_NO_AUTOSTART))
		return;
	ret = generic_com_on(hss, IPPROTO_TCP);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
}

/**
 * The HTTP sender.
 *
 * This is the only sender which does not FEC-encode the stream. This is not
 * necessary because HTTP sits on top of TCP, a reliable transport which
 * retransmits lost packets automatically. The sender employs per-client queues
 * which queue chunks of audio data if they can not be sent immediately because
 * the write operation would block. Most methods of the sender are implemented
 * as wrappers for the generic functions defined in \ref send_common.c.
 */
const struct sender http_sender = {
	.name = "http",
	.init = http_send_init,
	.shutdown = http_shutdown,
	.pre_select = http_pre_select,
	.post_select = http_post_select,
	.send = http_send,
	.shutdown_clients = http_shutdown_clients,
	.client_cmds = {
		[SENDER_on] = http_com_on,
		[SENDER_off] = http_com_off,
		[SENDER_deny] = http_com_deny,
		[SENDER_allow] = http_com_allow,
	},
	.help = generic_sender_help,
	.status = http_status,
};
