/* Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file user_list.c User handling for para_server. */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "error.h"
#include "crypt.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "user_list.h"

static INITIALIZED_LIST_HEAD(user_list);

/*
 * Wrapper for fgets(3).
 *
 * Unlike fgets(3), an integer value is returned. On success, this function
 * returns 1. On errors, -E_FGETS is returned. A zero return value indicates an
 * end of file condition.
 */
static int xfgets(char *line, int size, FILE *f)
{
again:
	if (fgets(line, size, f))
		return 1;
	if (feof(f))
		return 0;
	if (!ferror(f))
		return -E_FGETS;
	if (errno != EINTR) {
		PARA_ERROR_LOG("%s\n", strerror(errno));
		return -E_FGETS;
	}
	clearerr(f);
	goto again;
}

/**
 * Remove all entries from the user list.
 *
 * This is called on shutdown and when the user list is reloaded because the
 * server received SIGHUP.
 */
void user_list_deplete(void)
{
	struct user *u, *tmpu;

	list_for_each_entry_safe(u, tmpu, &user_list, node) {
		list_del(&u->node);
		free(u->name);
		free_public_key(u->pubkey);
		free(u);
	}
}

/**
 * Initialize the list of users allowed to connect to para_server.
 *
 * \param user_list_file The file containing access information.
 *
 * If this function is called for the second time, the contents of the
 * previous call are discarded, i.e. the user list is reloaded.
 *
 * This function either succeeds or calls exit(3).
 */
void user_list_init(const char *user_list_file)
{
	int ret = -E_USERLIST;
	FILE *file_ptr = fopen(user_list_file, "r");
	struct user *u;

	if (!file_ptr)
		goto err;

	user_list_deplete();
	for (;;) {
		int num;
		char line[255];
		/* keyword, name, key, perms */
		char w[255], n[255], k[255], p[255], tmp[4][255];
		struct asymmetric_key *pubkey;

		ret = xfgets(line, sizeof(line), file_ptr);
		if (ret <= 0)
			break;
		if (sscanf(line,"%200s %200s %200s %200s", w, n, k, p) < 3)
			continue;
		if (strcmp(w, "user"))
			continue;
		PARA_DEBUG_LOG("found entry for user %s\n", n);
		ret = get_public_key(k, &pubkey);
		if (ret < 0) {
			PARA_NOTICE_LOG("skipping entry for user %s: %s\n", n,
				para_strerror(-ret));
			continue;
		}
		/*
		 * In order to encrypt len := CHALLENGE_SIZE + 2 * SESSION_KEY_LEN
		 * bytes using RSA_public_encrypt() with EME-OAEP padding mode,
		 * RSA_size(rsa) must be greater than len + 41. So ignore keys
		 * which are too short. For details see RSA_public_encrypt(3).
		 */
		if (ret <= CHALLENGE_SIZE + 2 * SESSION_KEY_LEN + 41) {
			PARA_WARNING_LOG("public key %s too short (%d)\n",
				k, ret);
			free_public_key(pubkey);
			continue;
		}
		u = para_malloc(sizeof(*u));
		u->name = para_strdup(n);
		u->pubkey = pubkey;
		u->perms = 0;
		num = sscanf(p, "%200[A-Z_],%200[A-Z_],%200[A-Z_],%200[A-Z_]",
			tmp[0], tmp[1], tmp[2], tmp[3]);
		PARA_DEBUG_LOG("found %i perm entries\n", num);
		while (num > 0) {
			num--;
			if (!strcmp(tmp[num], "VSS_READ"))
				u->perms |= VSS_READ;
			else if (!strcmp(tmp[num], "VSS_WRITE"))
				u->perms |= VSS_WRITE;
			else if (!strcmp(tmp[num], "AFS_READ"))
				u->perms |= AFS_READ;
			else if (!strcmp(tmp[num], "AFS_WRITE"))
				u->perms |= AFS_WRITE;
			else /* unknown permission */
				PARA_WARNING_LOG("ignoring unknown permission: %s\n",
					tmp[num]);
		}
		para_list_add(&u->node, &user_list);
	}
	fclose(file_ptr);
	if (ret >= 0)
		return;
err:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

/**
 * Lookup a user in the user list.
 *
 * \param name The name of the user.
 *
 * \return A pointer to the corresponding user struct if the user was found, \p
 * NULL otherwise.
 */
const struct user *user_list_lookup(const char *name)
{
	const struct user *u;
	list_for_each_entry(u, &user_list, node) {
		if (strcmp(u->name, name))
			continue;
		return u;
	}
	return NULL;
}
